#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 17:45:10 2019

@author: felix and jonas
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D #Braucht man für 3D Plots
from scipy import linalg
from scipy.linalg import norm


class Vertex:
    def __init__(self, coords = np.zeros(3), edge = None, nu = None):
        self.edge = edge
        self.coords = coords
        self.nu = nu
        
    def __str__(self):
        return str(self.coords)
        
    def numofEdges(self):
        he = self.edge
        #hcurr = he
        hnext = he.twinEdge.nextEdge
        count =1
        while not hnext is he:
            count = count +1
            #hcurr = hnext
            hnext = hnext.twinEdge.nextEdge
        return count
        
    def interiourAngles(self):
        he = self.edge
        hcurr = he
        hnext = he.twinEdge.nextEdge
        alpha=0
        if not he.twinEdge.face is None:
            alpha = he.twinEdge.angle()
        
        while not hnext is he:
            hcurr = hnext
            hnext = hnext.twinEdge.nextEdge
            if not hcurr.twinEdge.face is None:
                alpha = alpha + hcurr.twinEdge.angle()
            
        return alpha
    
    def isboundary(self):
        he = self.edge
        hcurr = he
        hnext = he.twinEdge.nextEdge
        if he.twinEdge.face is None:
            return True
        
        while not hnext is he:
            hcurr = hnext
            hnext = hnext.twinEdge.nextEdge
            if hcurr.twinEdge.face is None:
               return True
        return False
    
    def getnu(self):
        return self.nu
        
        
        
        
        
class Face:
    def __init__(self, halfedge = None, normal = None):
        self.edge = halfedge
        self.normal = normal
    
    def __str__(self):
        string = "The face containes the edge: "+ str(self.edge.edgeVector())
        return string
        
    def setnormal(self):
        if self.isplanar():    
            e1 = self.edge.edgeVector()
            e2 = self.edge.nextEdge.edgeVector()
            v = np.cross(e1,e2)
            v = v / norm(v)
            self.normal = v
        else:
            print("the face is not planar")
        
    def numedges(self):
        count =1
        he= self.edge
        hecurr = he.nextEdge
        count =count +1
        while not hecurr.nextEdge is he:
            hecurr = hecurr.nextEdge
            count = count +1
        return count    
        
                
    def isplanar(self):
        b = True
        he = self.edge
        he1 = he
        he2 = he1.nextEdge
        he3 = he2.nextEdge
        a = np.array([he1.edgeVector(), he2.edgeVector(), he3.edgeVector()])
        det = np.linalg.det(a)
        if det is 0:
            b= False
            #print("false")
            return b
        while not he3 is he:
            he1 = he2
            he2 = he3
            he3 = he3.nextEdge
            a = np.array([he1.edgeVector(), he2.edgeVector(), he3.edgeVector()])
            det = np.linalg.det(a)
            if det is 0:
                b= False
                return b
        #print("true")
        return b  
        
    def getCenterofMass(self):
        n = self.numedges()
        if self.isplanar:
            he=self.edge
            v1=he.vertex.coords
            hcurr=he.nextEdge
            while not hcurr.nextEdge is he:
                v1 = v1+ hcurr.vertex.coords
                hcurr = hcurr.nextEdge
            v1 = (v1 + hcurr.vertex.coords)/n
            return v1
        print("the Face is not planar")
        return None

#we compute the diagonals of an planar quad and save as an 2*3 array.
#The first diagonal starts at the sourpoint of the face edge and the secound one at the target vertex of the face edge
    def diagonals(self):
        if self.isplanar() and self.numedges() is 4:
            
            h1 = self.edge
            p1=h1.sourcepoint().coords
            p2=h1.nextEdge.vertex.coords
            p3=h1.vertex.coords
            p4=h1.nextEdge.nextEdge.vertex.coords
            d1= p2-p1
            d2= p4-p3
            d=np.array([d1,d2])
            return d
            
    def dintersection(self):
        d = self.diagonals()
        if not d is None:
            e1=d[0]
            e2=d[1]
            e1=e1/norm(e1)
            e2=e2/norm(e2)
            e3= np.cross(e1,e2)
            A = np.array([e1,e2,e3])
            A = A.T
            p=self.edge.sourcepoint().coords - self.edge.vertex.coords
            x = np.linalg.solve(A, p)
            q = self.edge.vertex.coords + x[1]*e2
           
            return q
        else:
            print("die Facette ist kein ebens Viereck")  

#saves the signed distance from every vertex to the intersection pt. of the diagonals
    def sdistM(self):
        if self.isplanar and self.numedges() is 4:
            d=[]
            e= self.edge
            d.append(e.distanceM())
            d.append(e.nextEdge.distanceM())
            d.append(e.nextEdge.nextEdge.distanceM())
            d.append(e.nextEdge.nextEdge.nextEdge.distanceM())    
            return d
        else:
            return None


#compute the dual of a quad and save the value of nu at the vertices

    







class Halfedge:
    def __init__(self, nextEdge = None, twinEdge = None, point = None, face = None):
        self.nextEdge = nextEdge #Naechste Kante
        self.twinEdge = twinEdge #Zwillingskante
        self.vertex = point #Der Punkt auf den die Kante zeigt
        self.face = face #Zugehoerige Flaeche
    
            
#Give a string that says which two points are connected by the edge
    
    def __str__(self):
        edges = "The edge points from "+str(self.sourcepoint().coords)+" to "+str(self.vertex.coords)
        return edges
        


#Gives the sourcepoint of the halfedge using the twinEdge
    
    def sourcepoint(self):
        if self.twinEdge is None:
            hcurr = self.nextEdge
            hnext = hcurr.nextEdge
            #Continue until the next edge is the initial edge
            count = 0
            while(not hnext is self and count < 100):
                hcurr = hnext
                hnext = hcurr.nextEdge
                count += 1 #Just for savety reasons...
            #Return the point of the current edge
            return hcurr.vertex
        else:
            return self.twinEdge.vertex

    
#Gives the edgevector as an np.array of the halfedge using the twinEdge
    
    def edgeVector(self):
        v1 = self.vertex.coords
        v2 = self.sourcepoint().coords
        return v1-v2


#Gives the previous edge of the halfedge by iterating ofer the face and using nextEdge. This works for halfedges without twinEdges
#and is used in the methode that creates a halfedges mesh from an obj file
    
    def previousEdge(self):
        hcurr = self.nextEdge
        count = 0
        while(not hcurr.nextEdge is self and count < 100):
            hcurr = hcurr.nextEdge
            count += 1
        return hcurr


#Gives length of the halfedge
    
    def length(self):
        l = norm(self.edgeVector())
        return l
    
#Gives the angle that is enclosed by the Edge and its next Edge
    
    def angle(self):
        if self.face is None:
            print("boundary edge")
            v1 = self.edgeVector()
            v1 = v1/norm(v1)
            v2 = self.nextEdge.edgeVector()
            v2 = v2/norm(v2)
            a = np.dot(v1,v2)
            alpha = np.arccos(a)+np.pi
        else:
            v1 = self.edgeVector()
            v1 = -v1/norm(v1)
            v2 = self.nextEdge.edgeVector()
            v2 = v2/norm(v2)
            a = np.dot(v1,v2)
            alpha = np.arccos(a)
        return alpha
    
#Computes the discrete mean curvature of the edge   
    
    def meancurvature(self):
        if not self.isboundary() and not self.twinEdge.isboundary():
            f1= self.face
            f2= self.twinEdge.face
            c1=f1.getCenterofMass()
            c2=f2.getCenterofMass()
            f1.setnormal()
            f2.setnormal()
            n1 = f1.normal
            n2 = f2.normal
            d=c2-c1
            if abs(np.dot(n1,n2)) > 1:
                if np.dot(n1,n2) < 0:
                    theta = np.pi
                else:
                    theta=0
          
            else:
                theta = np.arccos(np.dot(n1,n2))
            
            if norm(d-n1+n2) < norm(n1+d-n2):
                theta = -theta
            
            l = self.length()
            H = 0.5*l*theta
            return H
        else:
            #print("this is a boundary edge")
            return 0.0
            
#Tests wether the edges belongs to the boundary
                      
    def isboundary(self):
        if self.face is None:
           return True

        else:
           return False         
                       
                       
               
    def distanceM(self):
        if self.face is None:
            print("boundary edge")
        else:
            f = self.face
            if f.isplanar() and f.numedges() is 4:
                M = f.dintersection()
                p = self.vertex.coords
                l = norm(M-p)
                i=0
                e = self
                while not e is self.face.edge:
                    e= e.nextEdge
                    i=i+1
                if i==0:    
                    return l
                if i==1:    
                    return -l
                if i==2:    
                    return -l
                if i==3:    
                    return l
            else:
                print("Die Facette ist kein ebens Viereck")
                       
                       
                       
class Mesh:
    def __init__(self):

        self.vertices = []
        self.faces = []
        self.edges = []
            
    def str2(self, edge = None):
        edges = "The edge " + str(self.edgenum(edge)) +" points from " + str(self.vertexnum(edge.sourcepoint())) + " to " + str(self.vertexnum(edge.vertex))
        return edges
                        
 
#---------------------------------------------------------------------------------------------------------------------                                                 
# Die Plot Methode für meshes:

                                                                                      
    def plot(self, color = 'b', points = True, edges = True, center = False, normals = False, pointnum = False, facenum =True, meancurvature = False, gausscurvature = False, diagonals = False, dintersection = False):
        figure = plt.figure()
        ax = figure.add_subplot(111, projection='3d')
        ax.set_aspect(1)
        
        if gausscurvature:
            
            #Wir wollen die Punkte des Meshes jetzt plotten und entsprechend ihrer Gausskrümmung einfärben. Dafür holen wir uns die Krümmung K als array
            #und wählen eine colormap die den Wert von K in eine Farbe übersetzt. Für diese Colormap können wir Grenzen fest legen (vmin, vmax)
            # mögliche cmaps sind :  'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic', 'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
            # 'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia','terrain', 'gist_stern',
            #'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
            #'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', etc...
            
            K = self.Gausscurvature()
            kmax = np.abs(max(K))
            kmin = np.abs(min(K))
            v= max(kmax,kmin)
            ax.scatter3D(self.coords()[:,0], self.coords()[:,1], self.coords()[:,2], s=100,  c = K, cmap = 'coolwarm', vmin=-v, vmax=v)
        else:
             if points: 
                  ax.scatter3D(self.coords()[:,0], self.coords()[:,1], self.coords()[:,2], color = "red")  
                  
        if pointnum:
            n = 0
            for v in self.vertices:
                ax.text(self.coords()[n,0]+0.05, self.coords()[n,1],self.coords()[n,2],n, fontsize=9)
                n = n+1
                
        if facenum :
              c = self.centerofFaces()
              n = 0
              for face in self.faces:  
                if face.numedges() is 4 and face.isplanar():
                     ax.text(c[n,0],c[n,1],c[n,2],n, fontsize=9)  
                     n = n+1
                      
        if meancurvature:
            H = self.meancurvature()
            maxh = max(H)
            minh = min(H)
            hmax = np.abs(maxh)
            hmin = np.abs(minh)
            
            v= max(hmax,hmin)
            if maxh-minh < 0.1:
                v=maxh+0.1
            #print(hmax)
            #print(v)
            norma = plt.Normalize(vmin = -v , vmax= v, clip=True)
            mapper = plt.cm.ScalarMappable(norm=norma, cmap=plt.cm.coolwarm)
            for face in self.faces:
                edge = face.edge
                n = self.edgenum(edge.nextEdge)
                #Die Punkte fuer die erste Kante waehlen:
                v0 = edge.vertex 
                v1 = edge.nextEdge.vertex
                #Die erste Kante plotten:
                ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color=mapper.to_rgba(H[n]))
                #
                
                #Eine Kante weiter gehen:
                edge = edge.nextEdge
                v0 = v1
                v1 = edge.nextEdge.vertex
                while (not v0 is face.edge.vertex): #Ueberpruefen, ob wir bei der Anfangskante angelangt sind.
                    #Kante plotten:
                    n=self.edgenum(edge)
                    ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color=mapper.to_rgba(H[n]))
                    #Eine Kante weiter gehen:
                    edge = edge.nextEdge
                    v0 = v1
                    v1 = edge.vertex
            
        else:
            if edges:
                    for face in self.faces:
                        edge = face.edge
                        #Die Punkte fuer die erste Kante waehlen:
                        v0 = edge.vertex 
                        v1 = edge.nextEdge.vertex
                        #Die erste Kante plotten:
                        ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color = color)
            
                        #Eine Kante weiter gehen:
                        edge = edge.nextEdge
                        v0 = v1
                        v1 = edge.nextEdge.vertex
                        while (not v0 is face.edge.vertex): #Ueberpruefen, ob wir bei der Anfangskante angelangt sind.
                            #Kante plotten:
                            ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color = color)
                            #Eine Kante weiter gehen:
                            edge = edge.nextEdge
                            v0 = v1
                            v1 = edge.vertex
            
        if diagonals:
            for face in self.faces:
                if face.numedges() is 4:    
                    d = face.diagonals()    
                    p2=face.edge.vertex.coords
                    p1=face.edge.sourcepoint().coords
                    ax.plot3D([p1[0],p1[0]+d[0][0]],[p1[1],p1[1]+d[0][1]],[p1[2],p1[2]+d[0][2]], color = "red")   
                    ax.plot3D([p2[0],p2[0]+d[1][0]],[p2[1],p2[1]+d[1][1]],[p2[2],p2[2]+d[1][2]], color = "red")   
                    
                    
        if dintersection:
            for face in self.faces:
                if face.numedges() is 4: 
                    M = face.dintersection()
                    M = np.array(M)
                    ax.scatter3D(M[0], M[1], M[2], color = 'g')      
                                          
        if center:
            c = self.centerofFaces()
            ax.scatter3D(c[:,0], c[:,1], c[:,2], color = 'g')
                
        if normals:
                
            c = self.centerofFaces()
            self.setnormals()
            N = self.getnormals()
            n = self.numfaces()
            for i in range(n):  
                ax.quiver(c[i,0], c[i,1], c[i,2], N[i,0], N[i,1], N[i,2],length=0.9, normalize=True)
                
        ax.set_xlabel('X axis')
        ax.set_ylabel('Y axis')
        ax.set_zlabel('Z axis')
       
                
                    
                            
        plt.show()
        return figure
        
    
    #Small Functions to obtain the lenght of the lists or the position of an object in the list:-------------------------------------------------------------------------------------
    #
    
    #Gives the number of faces of the mesh
    def numfaces(self): return len(self.faces)
    
    #Gives the number of vertices of the mesh
    def numvertices(self): return len(self.vertices)
        
    #Gives the number of halfedges of the mesh
    def numedges(self): return len(self.edges)
    
    #Return the number of a vertex, face or edge:
    def vertexnum(self, vertex):
        return self.vertices.index(vertex)
        
    def edgenum(self, edge):
        return self.edges.index(edge)
        
    def facenum(self, face):
        return self.faces.index(face)
    
    
#----------------------------------------------------------------------------
    def coords(self):
        #Zuerst Liste anlegen:
        coords = []
        for v in self.vertices:
            coords.append(v.coords)
        coords = np.array(coords)
        return coords
    
        
    def centerofFaces(self):
        c = []
        for f in self.faces:
            c.append(f.getCenterofMass())        
        c = np.array(c).astype(float)
        #print(c)
        return c
    
    def edges2(self):
        
        print( "List of edges: ")
        for e in self.edges:
            #edges = edges + " : from point " + str(self.vertexnum(e.sourcepoint()))+" " + str(e.sourcepoint().coords)+" to the point " + str(self.vertexnum(e.vertex))+" "+str(e.vertex.coords)+", "
            print( "the edge" + str(self.edgenum(e)) + " goes from point " + str(self.vertexnum(e.sourcepoint()))+" to the point " + str(self.vertexnum(e.vertex)))
        return "Ende"  
     
    def edgeVectors(self):
        E=[]
        for e in self.edges:
            E.append(e.edgeVector())
        return E     
    
    
    def setnormals(self):
        for f in self.faces:
            f.setnormal()
        
            
    def getnormals(self):
         N = []
         for f in self.faces:
             N.append(f.normal)
         N = np.array(N).astype(float)
         #print(str(N))
         return N
 
#Funktions for planar quad meshes to compute the diagonals, their interesctionpoints and the ratio of the diagonals
 
    def getdiagonals(self):
        dia = []
        for f in self.faces:
           dia.append(f.diagonals())
        return dia                                               
     
    def getDintersection(self):
        M = []
        for f in self.faces:
           M.append(f.dintersection())
        return M 
#Gives a list with the signed distance of the vertex to M, the intersection point of the diagonals                             
    def diadist(self):
        d=[]
        for e in self.edges:
            d.append(e.distanceM())                    
        return d                 
                                 
                                                 
 #Methoden zur Berechnung der Krümmungen einer Fläche:  ---------------------------    
    
    def Gausscurvature(self):
        K = []
        for v in self.vertices:
           K.append((2*np.pi - v.interiourAngles()))    
        return K
    
    def meancurvature(self):
        H=[]
        for e in self.edges:
            H.append(e.meancurvature())
        H = np.array(H) 
        return H
            
    
    def TotalGausscurvature(self):
        K=0
        for v in self.vertices:
           K = K + 2*np.pi - v.interiourAngles()
        K=np.array(K)    
        return K
 
             
    def Euler(self):
        E = self.numedges()/2
        V = self.numvertices()
        F = self.numfaces()
        X = V - E + F
        return X                       
                                    
                                                                                                                                         
# -----------------------------------------------------------------        
    
    #Eine Obj-Datei einlesen:
    def halfedgeFromObj(self, Dateipfad):
        #Initialzierung der Obj-Listen
        V = []
        F = []
        #Initialisierung der halfedge-Listen
        vertices = []
        faces = []
        halfedges = []
        #Auslesen des obj files
        file = open(Dateipfad, 'r')
        for line in file:
            if line.startswith('v '):
                V.append(line.split()[1:])
            if line.startswith('f '):
                f = line.split()[1:]
                for i in range(len(f)): 
                    f[i] = f[i].split('/')[0] #Only use the vertex coordinates of f 1/2/3 4/5/6 17/8/9
                F.append(f)
        V = np.array(V).astype(float)
        #F = np.array(F).astype(int)
        #print(F)
        #print(V)
        k = len(V)

        connect = np.ones( (k,k) )
        connect = -1*connect

        #safe the points as vertices in the vertices list
        for v in V:
            ver = Vertex(v,None)
            vertices.append(ver)

        #iterate over F to create and save the faces and halfedges. at first the halfedges do not have an opposite edge
        for f in F:
            #write an exception if the face is not a triangle
            #if len(f) != 3:
            #    print("this is not a triangle")
            n = len(f) #Number of vertices in face f
            
            #get the current position in the list of faces and halfedges
            lh = len(halfedges)
            lf = len(faces)

            #create the new face and add it to the list
            face = Face()
            faces.append(face)

            #create new halfedges, without opposite edge and next edge:
            for i in range(n):
                halfedges.append(Halfedge(None,None,vertices[int(f[i])-1],faces[lf])) # -1 because obj files start counting by 1
            
            #Connect the inner halfedges:
            for i in range(n):
                halfedges[lh +i].nextEdge = halfedges[lh + (i+1)%n]

            #Assign the out going edge to the points:
            for i in range(n):
                vertices[int(f[i])-1].edge = halfedges[lh + (i+1)%n]

            #Assign one of the halfedges to the face
            faces[lf].edge = halfedges[lh]

            # if the halfedge points from the vertex i to the vertex j we save it in the connect matrix at the point (i,j)
            for i in range(n):
                connect[int(f[(i-1)%n])-1, int(f[i])-1] = lh + i

        #Now we set the oposit edges for the interior edges
        for i in range(k):
            for j in range(i,k):
                if connect[i,j] > -1 and connect[j,i] > -1 :
                        halfedges[int(connect[i,j])].twinEdge = halfedges[int(connect[j,i])]
                        halfedges[int(connect[j,i])].twinEdge = halfedges[int(connect[i,j])]

        #Now we create the oposit edges for the exterior edges and add them to the list
        for h in halfedges:
            #first check if h is an exteriour edge,i.e. h has no twin
            if  h.twinEdge is None :
                #create a new halfedge that points on the source of h. The new edge does not have a face because it is an outer boundary edge
                #We cannot assign the nextEdge yet because it does not exists so far.
                halfedges.append(Halfedge(None,h,h.sourcepoint(),None))

                #save the position of the new halfedge such that we can later define its next edge and define the counter curr
                pos = len(halfedges)-1
                curr = len(halfedges)-1

                #assign the new edge to h as its twinEdge
                h.twinEdge = halfedges[pos]
                #Now we walk along the boundary and create the missing edges. Therefore we consider the previous edge of h and check if its part of the boundary.
                hp = h.previousEdge()
                while( not hp is h):
                    #if hp is part of the boundary we create its twin edge
                    if hp.twinEdge is None :
                        halfedges.append(Halfedge(None,hp,hp.sourcepoint(),None))
                        halfedges[curr].nextEdge = halfedges[curr+1]
                        curr = curr+1
                        hp.twinEdge = halfedges[curr]
                        hp = hp.previousEdge()
                    #if hp is not part of the boundary we continue with next outgoing edge in the vertexstar
                    else:
                        hp = hp.twinEdge.previousEdge()
                #know we can close the circle and give the boundary edge that we created first its nextEdge
                if hp is h:
                    halfedges[curr].nextEdge = halfedges[pos]

        #Finally we can assigne the new lists to the mesh
        self.vertices = vertices
        self.faces = faces
        self.edges = halfedges
        return self
        
        
        
#----------------------------------------------------------------------
# Ein Mesh als Obj Datei exportieren

    def exportobj(self, Dateipfad):
        
        file = open(Dateipfad,"w")
        v= len(self.vertices) 
        f= len(self.faces) 
        h= len(self.edges)/2
        file.write("# Number of vertices: " + str(v) +"\n" )
        file.write("# Number of faces: " + str(f) +"\n" ) 
        file.write("# Number of edges: " + str(h) +"\n" )        
        
        
        for p in self.vertices:
            file.write("v " + str(p.coords[0]) + " " + str(p.coords[1]) + " " + str(p.coords[2]) +"\n")
   
   
        for f in self.faces:
            file.write("f ")
            he = f.edge
            v = self.vertexnum(he.vertex)+1
            file.write(str(v))
            hcurr = he.nextEdge
            while hcurr != he:
                v = self.vertexnum(hcurr.vertex) +1
                file.write(" "+str(v))
                hcurr = hcurr.nextEdge  
                
            file.write("\n")
        file.close()
 
#-----------------------------------------------------------------------------------------------
#Functions to transform the mesh (Möbiustransformations)
   
   
    def scale(self,s=2):
         for v in self.vertices:
             v.coords = s*v.coords

    def translate(self, vektor = np.array([1,1,1])):
        for v in self.vertices:
             v.coords = vektor + v.coords
          
    #Spiegelung an der Einheitskugel         
                   
    def inverse(self):
        for v in self.vertices:
             if norm( v.coords) is 0 :
                 i=v.vertexnum()
                 print("Der Punkt " + str(i) + " wird nach unendlich abgebildet")
             else:
                v.coords = v.coords/ pow(norm( v.coords),2)
 
 
    def inverse2(self, center = np.array([0,0,0]), radius = 1):
            if not radius >0:
                print("Der Radius muss eine positive Zahl sein!")
            self.translate(-center)
            self.inverse()
            self.translate(center)
            self.scale(radius*radius)
 
 
# computes the koenigsdual for a given koenigsnet. This method only works if you write the methode dual in the class faces, that computes the dual of a plane quad and assigsns the values of the function nu to the vertices of the face.
            
            
    def koenigsdual2(self, startface = 0, startedge = 0):
         E= self.edgeVectors()

         f = self.faces[startface]
         e = f.edge
         A = self.diadist()
         #after we got he right face, we have to compute the halfedge e that gives us the direction for the next face.
         for i in range(startedge):
             e = e.nextEdge
             
         #We compute the dual of the first face:
         f.dual()
         #now we choose the next edge and go to all the faces in the row of this edge
         j=0
         d = e.nextEdge
         facei= startface
         while not d.twinEdge.face is None and not self.facenum(d.twinEdge.face) is facei:
                    f = d.twinEdge.face
                    
                    #get the nus, that were computed for the last face and the signed distance to the intersection of the diagonals of the original surface
                    nu4 = d.twinEdge.vertex.nu
                    nu1 = d.twinEdge.sourcepoint().nu
                    
                    a4 = A[self.edgenum(d.twinEdge)]
                    a1 = A[self.edgenum(d.twinEdge.previousEdge())]
                    a3 = A[self.edgenum(d.twinEdge.nextEdge)]
                    a2 = A[self.edgenum(d.twinEdge.nextEdge.nextEdge)]
                    #compute the nus for the vertices opposite to the edge d.twinEdge
                    nu2 = a2*nu4/a4
                    nu3 = a3*nu1/a1
                   
                    d.twinEdge.nextEdge.vertex.nu = nu3
                    d.twinEdge.nextEdge.nextEdge.vertex.nu = nu2
                    #using the nus we compute the new coordinates
                    
                    e1 = -E[self.edgenum(d.twinEdge.nextEdge)]
                    e2 = E[self.edgenum(d.twinEdge.previousEdge())]
                    d.twinEdge.nextEdge.vertex.coords = d.twinEdge.vertex.coords + 1/(nu4*nu3)*e1
                    d.twinEdge.nextEdge.nextEdge.vertex.coords = d.twinEdge.sourcepoint().coords + e2/(nu2*nu1)
                    
                    
                    
                    d = d.twinEdge.nextEdge.nextEdge
                    j=j+1
                   
         i=0
         while not e.twinEdge.face is None and not self.facenum(e.twinEdge.face) is facei:
             f = e.twinEdge.face
                    
             #get the nus, that were computed for the last face and the signed distance to the intersection of the diagonals of the original surface
             nu4 = e.twinEdge.vertex.nu
             nu1 = e.twinEdge.sourcepoint().nu
             
             a4 = A[self.edgenum(e.twinEdge)]
             a1 = A[self.edgenum(e.twinEdge.previousEdge())]
             a3 = A[self.edgenum(e.twinEdge.nextEdge)]
             a2 = A[self.edgenum(e.twinEdge.nextEdge.nextEdge)]
             #compute the nus for the vertices opposite to the edge d.twinEdge
             nu2 = a2*nu4/a4
             nu3 = a3*nu1/a1
             
             e.twinEdge.nextEdge.vertex.nu = nu3
             e.twinEdge.nextEdge.nextEdge.vertex.nu = nu2
             #using the nus we compute the new coordinates
             
             e1 = -E[self.edgenum(e.twinEdge.nextEdge)]
             e2 = E[self.edgenum(e.twinEdge.previousEdge())]
             e.twinEdge.nextEdge.vertex.coords = e.twinEdge.vertex.coords + 1/(nu4*nu3)*e1
             e.twinEdge.nextEdge.nextEdge.vertex.coords = e.twinEdge.sourcepoint().coords + e2/(nu2*nu1)
             
             
             
             e = e.twinEdge.nextEdge.nextEdge
             d = e.nextEdge
             i=i+1
             j=0
             facej = self.facenum(f)
             while not d.twinEdge.face is None and not self.facenum(d.twinEdge.face) is facej:
                    f = d.twinEdge.face
                 
                    #get the nus, that were computed for the last face and the signed distance to the intersection of the diagonals of the original surface
                    nu4 = d.twinEdge.vertex.nu
                    nu1 = d.twinEdge.sourcepoint().nu
                    
                    a4 = A[self.edgenum(d.twinEdge)]
                    a1 = A[self.edgenum(d.twinEdge.previousEdge())]
                    a3 = A[self.edgenum(d.twinEdge.nextEdge)]
                    a2 = A[self.edgenum(d.twinEdge.nextEdge.nextEdge)]
                    #compute the nus for the vertices opposite to the edge d.twinEdge
                    nu2 = a2*nu4/a4
                    nu3 = a3*nu1/a1
                   
                    d.twinEdge.nextEdge.vertex.nu = nu3
                    d.twinEdge.nextEdge.nextEdge.vertex.nu = nu2
                    #using the nus we compute the new coordinates
                  
                    e1 = -E[self.edgenum(d.twinEdge.nextEdge)]
                    e2 = E[self.edgenum(d.twinEdge.previousEdge())]
                    d.twinEdge.nextEdge.vertex.coords = d.twinEdge.vertex.coords + 1/(nu4*nu3)*e1
                    d.twinEdge.nextEdge.nextEdge.vertex.coords = d.twinEdge.sourcepoint().coords + e2/(nu2*nu1)
                    d = d.twinEdge.nextEdge.nextEdge
                    j=j+1
                    
         
             
         
    def faceiteration(self, startface = 0, startedge = 0):
         l= self.numfaces()
         f = self.faces[startface]
         e = f.edge
         for i in range(startedge):
             e = e.nextEdge
         #Here we write what do to with the first face:
         L=[]
         L.append(self.facenum(f))
         j=1
         d = e.nextEdge
         facei= startface
         while not d.twinEdge.face is None and not self.facenum(d.twinEdge.face) is facei:
                    f = d.twinEdge.face
                    L.append(self.facenum(f))
                    d = d.twinEdge.nextEdge.nextEdge
                    j=j+1
         i=1
         while not e.twinEdge.face is None and not self.facenum(e.twinEdge.face) is facei:
             f = e.twinEdge.face
             L.append(self.facenum(f))
             e = e.twinEdge.nextEdge.nextEdge
             d = e.nextEdge
             i=i+1
             j=1
             facej = self.facenum(f)
             while not d.twinEdge.face is None and not self.facenum(d.twinEdge.face) is facej:
                    f = d.twinEdge.face
                    L.append(self.facenum(f))
                    d = d.twinEdge.nextEdge.nextEdge
                    j=j+1
         if len(L)== l:
             print("all faces visited")
         else:
             print(len(L)-l)
         return L
             
            
        
class Quad(Mesh):
    
    
    
    def __init__(self, points = [[0,0,0],[1,0,0],[1,1,0],[0,1,0]]):
        Mesh.__init__(self) #Ruft den Mesh Initialisator auf.
        
        if type(points) == list:
            points = np.array(points)
        
        self.faces.append(Face())
        
        #Punkte anlegen
        for i in range(4):
            self.vertices.append(Vertex(points[i]))
        
        #Innen-Kanten anlegen:
        self.edges.append(Halfedge(None,None,self.vertices[0],self.faces[0]))
        self.edges.append(Halfedge(self.edges[0],None,self.vertices[3],self.faces[0]))
        self.edges.append(Halfedge(self.edges[1],None,self.vertices[2],self.faces[0]))
        self.edges.append(Halfedge(self.edges[2],None,self.vertices[1],self.faces[0]))
        self.edges[0].nextEdge = self.edges[3]
        
        #Jedem Punkt die ausgehende Innen-Kante zuordnen:
        self.vertices[0].edge = self.edges[3]
        self.vertices[1].edge = self.edges[2]
        self.vertices[2].edge = self.edges[1]            
        self.vertices[3].edge = self.edges[0]
        
        #Außenkanten setzen:
        for i in range(4):
            self.edges.append(Halfedge(None,self.edges[i],self.vertices[3-i],None))
            
        #Aussenkanten verknuepfen:
        for i in range(4):
            self.edges[i+4].nextEdge = self.edges[(i+1)%4+4] #Wir rechnen modulo n um im Kreis zu gehen.
                
        #Zwillingskanten zuordnen:
        for i in range(4):
            self.edges[i].twinEdge = self.edges[i + 4]
        
        #Dem face eine Halfedge zuordnen:
        self.faces[0].edge = self.edges[0]

############################## Test Area ###########################################

mesh1 = Mesh().halfedgeFromObj(r'/home/jtervooren/lehre/Tu Wien/Angewandte Geometrie/obj_files/ellipsoid.obj')
mesh2 = Mesh().halfedgeFromObj(r'/home/jtervooren/lehre/Tu Wien/Angewandte Geometrie/obj_files/ellipsoid.obj')

mesh1.plot(color = "b",points = True, edges = True, pointnum = False, facenum = True, normals = False, center = False, meancurvature = True, gausscurvature = True, diagonals = False, dintersection = False)
print(mesh1.faceiteration(0,2))


mesh1.koenigsdual2(0,2)

mesh1.plot(color = "b", points = True, edges = True, pointnum = False, facenum = True, normals = False, center = False, meancurvature = True, gausscurvature = True, diagonals = False, dintersection = False)



